(*  Title:      Config.thy
    ID:         $Id: Config.thy,v 1.3 2009/03/24 17:07:58 flokam Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Configurations, i.e. parallel activities for a functional
                ASP, based on the sigma-calculus and basic activities 
                defined in ASP.thy
                Def of configurations and reductions
      cf. Well-Formedness.thy for proof of well-formedness of the reduction presented here
   Name convention: 
               m??? = futmap == mapping fut-> term
               f??? = future id
               a, b = terms
               c??? = config
               A, B, G = activity
               E = contexts
*)

header {* Configurations for ASP *}

theory Config imports ContextLocalRed Infinite_Set begin

declare cCompose_and_Fill[simp]

(* noFV is the same as Sigma.closed *)

(* noRef is true for a term containing no future and no active object reference, 
   i.e. for a source term *)
primrec 
 noRef        :: "sterm \<Rightarrow> bool"
and
 noRef_option :: "sterm option \<Rightarrow> bool"
where
   noRef_Bvar  : "noRef (Bvar n) = True"
 | noRef_Fvar  : "noRef (Fvar x) = True"
 | noRef_Act   : "noRef (Active a) = noRef a"
 | noRef_ActRef: "noRef (ActRef a) = False"
 | noRef_FutRef: "noRef (FutRef a) = False"
 | noRef_Call  : "noRef (Call a l b) = ((noRef a) \<and> (noRef b))"
 | noRef_Upd   : "noRef (Upd a l b) = ((noRef a) \<and> (noRef b))"
 | noRef_Obj   : "noRef (Obj f T)  = (\<forall>l\<in>dom f. noRef_option (f l))" 
 | noRef_None  : "noRef_option (None) = True"
 | noRef_Some  : "noRef_option (Some a) = noRef a"

types futmap = "FutureRef ~=> sterm"
      configuration = "ActivityRef ~=> (futmap * sterm)"

subsection{*Configuration Accessors*}

(* we define operators that access the active object, the request queue 
or a single request of a given activity *)
constdefs ActiveObject :: "[configuration, ActivityRef] \<Rightarrow> sterm" (infixl ".\<^sub>A" 50)
  "(ActiveObject c A) == (snd (the (c A)))"
constdefs RequestQueue :: "[configuration, ActivityRef] \<Rightarrow> futmap" (infixl ".\<^sub>R" 50)
  "(RequestQueue c A) == (fst (the (c A)))"
constdefs OneRequest :: "[configuration, ActivityRef, FutureRef] \<Rightarrow> sterm" ("_.\<^sub>F_<_>" [50,0,0] 50)
  "(OneRequest c A fi) == (the ((fst (the (c A))) fi))"

lemma ActiveObjectE: "\<lbrakk> c.\<^sub>AA = a; A \<in> dom c \<rbrakk> \<Longrightarrow> \<exists>m. c A = Some (m,a)"
  unfolding ActiveObject_def by auto

lemma RequestQueueE: "\<lbrakk> c.\<^sub>RA = m; A \<in> dom c \<rbrakk> \<Longrightarrow> \<exists>a. c A = Some (m,a)"
  unfolding RequestQueue_def by auto

lemma OneRequestE: "\<lbrakk> c.\<^sub>FA<f> = s; A \<in> dom c; f \<in> dom(c.\<^sub>RA) \<rbrakk> 
  \<Longrightarrow> \<exists>m a. c A = Some (m,a) \<and> m f = Some s"
  unfolding OneRequest_def RequestQueue_def by auto

lemma ActiveObjectI: "c A = Some (m,a) \<Longrightarrow> c.\<^sub>AA = a"
  unfolding ActiveObject_def by auto

lemma RequestQueueI: "c A = Some (m,a) \<Longrightarrow> c.\<^sub>RA = m"
  unfolding RequestQueue_def by auto

lemma OneRequestI: "\<lbrakk> c A = Some (m,a); m f = Some s \<rbrakk> \<Longrightarrow> c.\<^sub>FA<f> = s"
  unfolding OneRequest_def by auto

lemma ActivityDef: "A \<in> dom c \<Longrightarrow> c A = Some (c.\<^sub>RA,c.\<^sub>AA)"
  unfolding ActiveObject_def RequestQueue_def by auto

subsection{*Reduction*}
(* now the semantics, see ASPfun papers for details *)
inductive zeta :: "[configuration, configuration] \<Rightarrow> bool"  (infixl "\<rightarrow>\<^sub>\<zeta>" 50)
  where
  zeta_local [simp, intro!]: "\<lbrakk> A \<in> dom c; fk \<in> dom (c.\<^sub>RA); c.\<^sub>FA<fk> \<rightarrow>\<^sub>\<beta> a' \<rbrakk> 
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> a'),c.\<^sub>AA))"
| active [simp, intro!]    : "\<lbrakk> G \<notin> dom c; A \<in> dom c; fk \<in> dom (c.\<^sub>RA); 
                                c.\<^sub>FA<fk> = (E::general_context)\<up>Active a; 
                                lc_ctxt E; lc a; closed a \<rbrakk>
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),c.\<^sub>AA))
                                        (G \<mapsto> (empty,a))"
| request [simp,intro]     : "\<lbrakk> \<forall>A\<in>dom c. fn \<notin> dom (c.\<^sub>RA); A \<in> dom c; 
                                 B \<in> dom c; A \<noteq> B; fk \<in> dom (c.\<^sub>RA);
                                c.\<^sub>FA<fk> = (E::general_context)\<up>Call (ActRef B) l p;
                                lc_ctxt E; lc p; closed p \<rbrakk>
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn),c.\<^sub>AA))
                                        (B \<mapsto> ((c.\<^sub>RB)(fn \<mapsto> Call (c.\<^sub>AB) l p),c.\<^sub>AB))"
| request_self[simp,intro] : "\<lbrakk> \<forall>A\<in>dom c. fn \<notin> dom (c.\<^sub>RA); A \<in> dom c; fk \<in> dom (c.\<^sub>RA); 
                                c.\<^sub>FA<fk> = (E::general_context)\<up>Call (ActRef A) l p;
                                lc_ctxt E; lc p; closed p \<rbrakk>
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn)
                                                     (fn \<mapsto> Call (c.\<^sub>AA) l p),c.\<^sub>AA))"
| update_AO [simp,intro]   : "\<lbrakk> G \<notin> dom c; A \<in> dom c; B \<in> dom c; fk \<in> dom (c.\<^sub>RA);
                                c.\<^sub>FA<fk> = (E::general_context)\<up>Upd (ActRef B) l p;
                                lc_ctxt E; body p; closed p \<rbrakk>
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),c.\<^sub>AA))
                                        (G \<mapsto> (empty, Upd (c.\<^sub>AB) l p))"
| reply   [simp, intro!]   : "\<lbrakk> A \<in> dom c; fk \<in> dom (c.\<^sub>RA); 
                                c.\<^sub>FA<fk> = (E::general_context)\<up>FutRef fi;
                                B \<in> dom c; fi \<in> dom (c.\<^sub>RB) \<rbrakk>
                             \<Longrightarrow> c \<rightarrow>\<^sub>\<zeta> c (A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>(c.\<^sub>FB<fi>)),c.\<^sub>AA))"

abbreviation
  beta_reds :: "[configuration, configuration] \<Rightarrow> bool"  (infixl "~->>" 50) where
  "s ~->> t == zeta^** s t"

notation (latex)
  beta_reds  (infixl "\<rightarrow>\<^sub>\<zeta>\<^sup>*" 50)

(* a conf is finite if it has a finite number of activities and requests *)
constdefs finite_conf :: "configuration \<Rightarrow> bool"
  "finite_conf c == finite (dom c) \<and> (\<forall>A\<in>dom c. finite (dom(c.\<^sub>RA)))"

lemma finite_step[rule_format]: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> finite_conf c \<longrightarrow> finite_conf c'"
  by (induct c c' rule: zeta.induct, 
      unfold finite_conf_def RequestQueue_def ActiveObject_def OneRequest_def, auto)

lemma finite_conf_rtrancl[rule_format]: "\<lbrakk> x \<rightarrow>\<^sub>\<zeta>\<^sup>* y \<rbrakk> \<Longrightarrow> finite_conf x \<longrightarrow> finite_conf y"
  by (erule rtranclp_induct, auto simp: finite_step)

(* if a set is finite, but its base type is infinite, then there is always a fresh
   element *)
lemma finite_set_subset_infinite_fresh: "finite (A::(('a::inftype)set)) \<Longrightarrow> \<exists>x::'a. x \<notin> A"
  by (auto simp: ex_new_if_finite[OF inftype_class.infinite])

(* We need to state that the domains of configurations are over infinite types.
   However, the types FutureRef and ActivityRef are already defined as datatypes
   embedding nat. Hence we can _derive_ the infinite of these types from there. *)

instance nat :: inftype
  by (intro_classes, rule Finite_Set.infinite_UNIV_nat)

lemma surj_fref: "surj fref"
  unfolding surj_def
  by (auto, case_tac y, rule_tac x = nat in exI, assumption)

lemma inj_fref: "inj fref"
  unfolding inj_on_def by simp

instance FutureRef :: inftype 
proof (intro_classes, rule_tac s = "range fref" and t = UNIV in subst)
  show "range fref = UNIV" by (rule surj_range, rule surj_fref)
next
  from range_inj_infinite[OF inj_fref] show "infinite (range fref)" by assumption
qed

(* Here is now the old axiom fresh_futref, but with the assumption that
   the set A (in the application of the lemma to be replaced by dom c)
   is finite *)
lemma fresh_futref_finite: "finite F \<Longrightarrow> \<exists>(id::FutureRef). id \<notin> F"
  by (erule finite_set_subset_infinite_fresh)


(* Same procedure for ActivityRefs *)
lemma surj_aref: "surj aref"
  unfolding surj_def
  by (auto, case_tac y, rule_tac x = nat in exI, assumption)

lemma inj_aref: "inj aref"
  unfolding inj_on_def by simp

instance ActivityRef :: inftype 
proof (intro_classes, rule_tac s = "range aref" and t = UNIV in subst)
  show "range aref = UNIV" by (rule surj_range, rule surj_aref)
next
  from range_inj_infinite[OF inj_aref] show "infinite (range aref)" by assumption
qed

lemma fresh_actref_finite: "finite A \<Longrightarrow> \<exists>(id::ActivityRef). id \<notin> A"
  by (erule finite_set_subset_infinite_fresh)

(* Here could come lemmata to infer finite(dom c) and
   (! A: dom c. finite(dom (fst c A)) from finite_conf.
   These are just normal elimination rules. *)
lemma finite_confE1: "finite_conf c \<Longrightarrow> finite (dom c)"
  by (simp add: finite_conf_def)

lemma finite_confE2: "\<lbrakk> finite_conf c; A \<in> dom c \<rbrakk> \<Longrightarrow> finite (dom (c.\<^sub>RA))"
  by (simp add: finite_conf_def)


(* 2. All reachable configs are finite: 
      We first define an initial config c0 (ie a single act only consisting in a single request, 
      mapped to a tern without future or activity ref) 
      c0 is provably finite. Using finite_conf_rtrancl we get that any c with c0 ~->> c 
      is finite. So assuming reachable configs we get our old freshness axioms with
      the help of fresh_fut/actref_finite. *)

constdefs 
  init_config :: "sterm \<Rightarrow> configuration"
  "init_config a == [aref 0 \<mapsto> ([fref 0 \<mapsto> a], Obj empty (Object empty))]"

(* more structure? 
  init_configs :: "configuration set"
    "init_configs == {c. \<exists>a. (noRef a) \<and> c = init_config a}"
*)

  reachable_configs :: " configuration set" 
  "reachable_configs == {c. \<exists>a. (noRef a) \<and> init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c}"

lemma finite_init_config: "finite_conf (init_config a)"
  unfolding finite_conf_def init_config_def RequestQueue_def by auto

lemma finite_reachable_configs: "\<forall>c\<in>reachable_configs. finite_conf c"
  unfolding reachable_configs_def
proof (intro ballI, simp)
  fix c :: configuration
  assume "\<exists>a. noRef a \<and> init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c"
  then obtain a where "init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c" by auto
  from finite_conf_rtrancl[OF this finite_init_config[of a]]
  show "finite_conf c" by assumption
qed

(* A simpler form that is needed *)
lemma finite_reachable_configsL: "c \<in> reachable_configs \<Longrightarrow> finite (dom c)"
  by (rule finite_confE1, simp add: finite_reachable_configs)

(* And one for the union of the request queues *)
lemma finite_conf_unionE: "finite_conf c \<Longrightarrow> finite (\<Union>{dom (c.\<^sub>RA) |A. A \<in> dom c})"
proof -
  have "(\<Union>{dom (c.\<^sub>RA) |A. A \<in> dom c}) \<subseteq> (\<Union>A\<in>dom c. dom (c.\<^sub>RA))" by force
  note finite_subset[OF this]
  moreover
  assume "finite_conf c"
  note finite_UN_I[OF finite_confE1[OF this] finite_confE2[OF this]]
  from this[of "\<lambda>x. x"]
  have "finite (\<Union>A\<in>dom c. dom (c.\<^sub>RA))" by simp
  ultimately show ?thesis by simp
qed

(*  From these we can then infer  the original axioms
axioms fresh_futref: " \<exists>(id::FutureRef). id \<notin>F"
       fresh_actref: " \<exists>(id::ActivityRef). id \<notin>A"

but now relative to the assumption that the configurations are reachable.
(See TypedASP)
*)

end
